/*
Copyright (C) 2001 StrmnNrmn

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include "stdafx.h"

#include "PSPRenderer.h"
#include "Texture.h"
#include "NativeTexture.h"
#include "RDP.h"
#include "GraphicsContext.h"
#include "PSPColour.h"

#include "DebugDisplayList.h"

#include "Debug/DBGConsole.h"

#include "Core/Memory.h"			// We access the memory buffers
#include "Core/ROM.h"

#include "SysPSP/DaedalusVFPU.h"

#include "Utility/Profiler.h"
#include "PngUtil.h"

#include <pspgu.h>
#include <pspgum.h>
#include <psputils.h>
#include <pspdebug.h>
#include <pspctrl.h>
#include <pspdisplay.h>

#include <vector>

#include <pspvfpu.h>

float DECAL_Z_OFFSET = +3.1f;		// Found through trial an error for the PSP

bool gSoftwareClipping( true );
bool gUseVFPUTnL( true );

#include "PushStructPack1.h"
struct FiddledVtx
{
	s16 y;
	s16 x;

	s16 flag;
	s16 z;

	s16 tv;
	s16 tu;

	union
	{
		struct
		{
			u8 rgba_a;
			u8 rgba_b;
			u8 rgba_g;
			u8 rgba_r;
		};
		struct
		{
			s8 norm_a;
			s8 norm_z;	// b
			s8 norm_y;	// g
			s8 norm_x;	// r
		};
	};
};

DAEDALUS_STATIC_ASSERT( sizeof(FiddledVtx) == 16 );


struct FiddledVtxDKR
{
	s16 y;
	s16 x;

	u8 a;
	u8 b;
	s16 z;

	u8 g;
	u8 r;
};

#include "PopStructPack.h"


extern "C"
{
void	_TransformVerticesWithLighting( const Matrix4x4 * world_matrix, const Matrix4x4 * projection_matrix, const FiddledVtx * p_in, const DaedalusVtx4 * p_out, u32 num_vertices, const v4 * p_ambient, const DaedalusLight * p_lights, u32 num_lights );
void	_TransformVerticesWithColour( const Matrix4x4 * world_matrix, const Matrix4x4 * projection_matrix, const FiddledVtx * p_in, const DaedalusVtx4 * p_out, u32 num_vertices );
u32		_ClipToHyperPlane( DaedalusVtx4 * dest, const DaedalusVtx4 * source, const v4 * plane, u32 num_verts );
}



// Bits for clipping
// +-+-+-
// xxyyzz
#define Z_NEG  0x01
#define Z_POS  0x02
#define Y_NEG  0x04
#define Y_POS  0x08
#define X_NEG  0x10
#define X_POS  0x20

// Test all but Z_NEG (for No Near Plane microcodes)
static const u32 CLIP_TEST_FLAGS( X_POS | X_NEG | Y_POS | Y_NEG | Z_POS );

//#define TEXRECT_USE_3D_TRANSFORM

#define GL_TRUE                           1
#define GL_FALSE                          0

#undef min
#undef max

enum CycleType
{
	CYCLE_1CYCLE = 0,		// Please keep in this order - matches RDP 
	CYCLE_2CYCLE,
	CYCLE_COPY,
	CYCLE_FILL,
};


static u32 sViWidth = 320;
static u32 sViHeight = 240;

static const u32 gPlaceholderTextureWidth( 16 );
static const u32 gPlaceholderTextureHeight( 16 );
static const float gFillRectDepth( 0.0f );
static const float gTexRectDepth( 0.0f );

static const float gMaxZValue( 1.0f);

u32		__attribute__((aligned(16)))	gWhiteTexture[gPlaceholderTextureWidth * gPlaceholderTextureHeight ];
u32		__attribute__((aligned(16)))	gPlaceholderTexture[gPlaceholderTextureWidth * gPlaceholderTextureHeight ];
u32		__attribute__((aligned(16)))	gSelectedTexture[gPlaceholderTextureWidth * gPlaceholderTextureHeight ];

//
//	If this flag is set, we override the diffuse colour (or just alpha) in the vertex with the specified colour
//


extern void HandleEndOfFrame();


void			DaedalusVtx4::Interpolate( const DaedalusVtx4 & lhs, const DaedalusVtx4 & rhs, float factor )
{
	ProjectedPos = lhs.ProjectedPos + (rhs.ProjectedPos - lhs.ProjectedPos) * factor;
	TransformedPos = lhs.TransformedPos + (rhs.TransformedPos - lhs.TransformedPos) * factor;
	color = lhs.color + (rhs.color - lhs.color) * factor;
	t0 = lhs.t0 + (rhs.t0 - lhs.t0) * factor;
}


//*****************************************************************************
// Creator function for singleton
//*****************************************************************************
template<> bool CSingleton< PSPRenderer >::Create()
{
	DAEDALUS_ASSERT_Q(mpInstance == NULL);
	
	mpInstance = new PSPRenderer();
	if (!mpInstance)
	{
		return false;
	}
	
	return true;
}

enum MuxVal
{
	MUX_0 = 0,
	MUX_1,
	MUX_COMBINED,
	MUX_TEXEL0,
	MUX_TEXEL1,
	MUX_PRIM,
	MUX_SHADE,
	MUX_ENV,
	MUX_COMBALPHA,
	MUX_T0_ALPHA,
	MUX_T1_ALPHA,
	MUX_PRIM_ALPHA,
	MUX_SHADE_ALPHA,
	MUX_ENV_ALPHA,
	MUX_LODFRAC,
	MUX_PRIMLODFRAC,
	MUX_UNK,
	MUX_PRIM_MOD_ENV,
	MUX_PRIM_ENV_MEAN
};

//*****************************************************************************
//
//*****************************************************************************
PSPRenderer::PSPRenderer() :
	m_fScreenMultX(2.0f),
	m_fScreenMultY(2.0f),

	m_bTextureEnable(false),
	m_bLightingEnable(false),

	m_dwNumLights(0),
	m_dwTextureFactor(MAKE_COLOUR_U32(255,255,255,255)),
	m_bZBuffer(false),
		
	m_bCullFront(false),
	m_bCullBack(true),

	m_dwAlpha(0xFF),

	mSmooth( true ),
	mSmoothShade( true ),
	
	m_fFogMult(0.0f),			// * 0 + 0 = 0
	m_fFogOffset(0.0f),
	m_bFogEnable(false),
	m_dwFogColor(0x00FFFFFF),

	m_bTextureGen(false),

	m_dwProjectionTop(0),
	m_dwModelViewTop(0),

	m_dwNumIndices(0),
	mVtxClipFlagsUnion( 0 ),

	mbModulateDiffuseColour( false ),
	mSetDiffuseMask( 0 ),
	mSetDiffuseColour( 0xffffffff ),

	mModulateDiffuseMask( 0 ),
	mModulateDiffuseColour( 0xffffffff ),

	m_dwNumTrisRendered( 0 ),
	m_dwNumTrisClipped( 0 )
#ifdef DAEDALUS_DEBUG_DISPLAYLIST
,	mRecordCombinerStates( false )
#endif
{
	for ( u32 t = 0; t < 2; t++ )
	{
		m_lpsTexture[t] = NULL;
		m_fTileWidth[t] = 32.0f;
		m_fTileHeight[t] = 32.0f;
		m_fTexWidth[t] = 32.0f;
		m_fTexHeight[t] = 32.0f;
		m_fInvTileWidth[t] = 1/32.0f;
		m_fInvTileHeight[t] = 1/32.0f;
		m_fInvTexWidth[t] = 1/32.0f;
		m_fInvTexHeight[t] = 1/32.0f;
	}

	memset( mLights, 0, sizeof(mLights) );

	m_fTexScaleX = 1.0f;
	m_fTexScaleY = 1.0f;

	mAmbientColour = v4( 1.0f, 1.0f, 1.0f, 1.0f );

	gRDPMux._u64 = 0;

	memset( gWhiteTexture, 0xff, sizeof(gWhiteTexture) );

	u32	texel_idx( 0 );
	const u32	COL_MAGENTA( MAKE_COLOUR_U32( 255, 0, 255 ) );
	const u32	COL_GREEN( MAKE_COLOUR_U32( 0, 255, 0 ) );
	const u32	COL_BLACK( MAKE_COLOUR_U32( 0,0,0 ) );
	for(u32 y = 0; y < gPlaceholderTextureHeight; ++y)
	{
		for(u32 x = 0; x < gPlaceholderTextureWidth; ++x)
		{
			gPlaceholderTexture[ texel_idx ] = ((x&1) == (y&1)) ? COL_MAGENTA : COL_BLACK;
			gSelectedTexture[ texel_idx ]    = ((x&1) == (y&1)) ? COL_GREEN   : COL_BLACK;

			texel_idx++;
		}
	}

	RestoreRenderStates();
}

//*****************************************************************************
//
//*****************************************************************************
PSPRenderer::~PSPRenderer()
{	
	for ( u32 i = 0; i < 2; i++ )
	{
		if ( m_lpsTexture[i] != NULL )
		{
			m_lpsTexture[i]->Release();
			m_lpsTexture[i] = NULL;
		}
	}
}

//*****************************************************************************
//
//*****************************************************************************
bool PSPRenderer::RestoreRenderStates()
{
	// Initialise the device to our default state

	// We do our own culling
	sceGuDisable(GU_CULL_FACE);

	// But clip our tris please
	sceGuEnable(GU_CLIP_PLANES);
	//sceGuDisable(GU_CLIP_PLANES);

	sceGuScissor(0,0, 480,272);
	sceGuEnable(GU_SCISSOR_TEST);

	// We do our own lighting
	sceGuDisable(GU_LIGHTING);

	sceGuAlphaFunc(GU_GEQUAL, 0x04, 0xff );
	sceGuEnable(GU_ALPHA_TEST);
	sceGuBlendFunc(GU_ADD, GU_SRC_ALPHA, GU_ONE_MINUS_SRC_ALPHA, 0, 0);
	sceGuEnable(GU_BLEND);

	// Default is ZBuffer disabled
	sceGuDepthMask(GL_TRUE);	// GL_TRUE to disable z-writes
	sceGuDepthFunc(GU_GEQUAL);		// GEQUAL?
	sceGuDisable(GU_DEPTH_TEST);

	// Initialise all the renderstate to our defaults.
	sceGuShadeModel(GU_SMOOTH);

	sceGuTexEnvColor(m_dwTextureFactor);
	sceGuTexOffset(0.0f,0.0f);

	//sceGuFog(near,far,m_dwFogColor);
	// Texturing stuff
	sceGuTexFunc(GU_TFX_REPLACE,GU_TCC_RGB);
	sceGuTexFilter(GU_LINEAR,GU_LINEAR);
	sceGuTexWrap(GU_REPEAT,GU_REPEAT);

	sceGumMatrixMode(GU_PROJECTION);
	sceGumLoadIdentity();

	sceGumMatrixMode(GU_VIEW);
	sceGumLoadIdentity();

	sceGumMatrixMode(GU_MODEL);
	sceGumLoadIdentity();

	return true;		
}

//*****************************************************************************
//
//*****************************************************************************
// Reset for a new frame
void	PSPRenderer::Reset()
{
	ResetMatrices();

	m_dwNumIndices = 0;
	mVtxClipFlagsUnion = 0;

	m_dwNumTrisRendered = 0;
	m_dwNumTrisClipped = 0;

	u32 dwScaleX = Memory_VI_GetRegister( VI_X_SCALE_REG ) & 0xFFF;
	u32 dwScaleY = Memory_VI_GetRegister( VI_Y_SCALE_REG ) & 0xFFF;

	float fScaleX = (float)dwScaleX / (1<<10);
	float fScaleY = (float)dwScaleY / (1<<10);

	//DBGConsole_Msg(DEBUG_VI, "VI_X_SCALE_REG set to 0x%08x (%f)", dwValue, 1/fScale);
	//DBGConsole_Msg(DEBUG_VI, "VI_Y_SCALE_REG set to 0x%08x (%f)", dwValue, 1/fScale);

	u32 dwHStartReg = Memory_VI_GetRegister( VI_H_START_REG );
	u32 dwVStartReg = Memory_VI_GetRegister( VI_V_START_REG );

	u32	hstart = dwHStartReg >> 16;
	u32	hend = dwHStartReg & 0xffff;
	//DBGConsole_Msg( 0, "h start/end %x %x", hstart, hend );		// 128 725 - 597

	u32	vstart = dwVStartReg >> 16;
	u32	vend = dwVStartReg & 0xffff;
	//DBGConsole_Msg( 0, "v start/end %x %x", vstart, vend );		// 56 501 - 445

	sViWidth  = (u32)( (hend-hstart)    * fScaleX);
	sViHeight = (u32)(((vend-vstart)/2) * fScaleY);				// 

	if ( sViWidth == 0 )
	{
		sViWidth = 320;
	}
	if ( sViHeight == 0 )
	{
		sViHeight = 240;
	}
}

//*****************************************************************************
//
//*****************************************************************************
void PSPRenderer::BeginScene()
{
	CGraphicsContext::Get()->BeginFrame();

	RestoreRenderStates();

	extern bool			gNeedZBufferUdate;
	gNeedZBufferUdate = true;

#ifdef DAEDALUS_DEBUG_DISPLAYLIST
	mRecordedCombinerStates.clear();
#endif
}

//*****************************************************************************
//
//*****************************************************************************
void PSPRenderer::EndScene()
{
	CGraphicsContext::Get()->EndFrame();

	HandleEndOfFrame();
}

//*****************************************************************************
//
//*****************************************************************************
void	PSPRenderer::SelectPlaceholderTexture( EPlaceholderTextureType type )
{
	switch( type )
	{
	case PTT_WHITE:			sceGuTexImage(0,gPlaceholderTextureWidth,gPlaceholderTextureHeight,gPlaceholderTextureWidth,gWhiteTexture); break;
	case PTT_SELECTED:		sceGuTexImage(0,gPlaceholderTextureWidth,gPlaceholderTextureHeight,gPlaceholderTextureWidth,gSelectedTexture); break;
	case PTT_MISSING:		sceGuTexImage(0,gPlaceholderTextureWidth,gPlaceholderTextureHeight,gPlaceholderTextureWidth,gPlaceholderTexture); break;
	default:
		DAEDALUS_ERROR( "Unhandled type" );
		break;
	}
}

//*****************************************************************************
//
//*****************************************************************************
void PSPRenderer::SetD3DViewport( u32 dwX, u32 dwY, u32 dwWidth, u32 dwHeight )
{
	m_VP.X = dwX;
	m_VP.Y = dwY;
	m_VP.Width = dwWidth;
	m_VP.Height = dwHeight;
	m_VP.MinZ = 0.0f;
	m_VP.MaxZ = 1.0f;

//	g_pD3DDev->SetViewport( &m_VP );

	m_fScreenMultX = (float)m_VP.Width / (float)(sViWidth);
	m_fScreenMultY = (float)m_VP.Height / (float)(sViHeight);
}

//*****************************************************************************
//
//*****************************************************************************
void PSPRenderer::SetN64Viewport( const v3 & scale, const v3 & trans )
{
	mVpScale = scale;
	mVpTrans = trans;

	mVpScale.x *= (float)m_fScreenMultX;
	mVpScale.y *= -(float)m_fScreenMultY;		// Apply the flip here to invert the y to the correct value
	mVpScale.z *= 1.0f / 256.0f;

	mVpTrans.x *= (float)m_fScreenMultX;
	mVpTrans.y *= (float)m_fScreenMultY;
	mVpTrans.z *= 1.0f / 256.0f;
}

//*****************************************************************************
//
//*****************************************************************************
void	PSPRenderer::InitCycleCopy( )
{
	DL_PF("CycleType: Copy");

	mSetDiffuseMask = 0;
	mSetDiffuseColour = 0xffffffff;
	mModulateDiffuseMask = 0;
	mModulateDiffuseColour = 0xffffffff;

	sceGuTexFunc(GU_TFX_REPLACE,GU_TCC_RGBA);
	if(m_lpsTexture[0] != NULL)
	{
		CNativeTexture *	p_texture( m_lpsTexture[0]->GetTexture() );
		if(p_texture != NULL)
		{
			p_texture->InstallTexture();
		}
	}
}

//*****************************************************************************
//
//*****************************************************************************
void	PSPRenderer::InitCycleFill()
{
	DL_PF("CycleType: Fill");

	mSetDiffuseMask = 0;
	mSetDiffuseColour = 0xffffffff;
	mModulateDiffuseMask = 0;
	mModulateDiffuseColour = 0xffffffff;

	// XXXX
	// Need to select diffuse colour here - ignore the texture!
	// Need to remove the texture 
	//sceGuDisable(GU_TEXTURE_2D);
	sceGuEnable(GU_TEXTURE_2D);
	sceGuTexMode(GU_PSM_8888,0,0,GL_TRUE);		// maxmips/a2/swizzle = 0
	sceGuTexImage(0,gPlaceholderTextureWidth,gPlaceholderTextureHeight,gPlaceholderTextureWidth,gWhiteTexture);
	sceGuTexFunc(GU_TFX_MODULATE,GU_TCC_RGB);
}

//*****************************************************************************
//
//*****************************************************************************
void PSPRenderer::PrepareVerticesForBlendMode( DaedalusVtx * p_vertices, u32 num_verts )
{
	DAEDALUS_PROFILE( "PSPRenderer::PrepareVerticesForBlendMode" );

	DAEDALUS_ASSERT( (mSetDiffuseMask & mModulateDiffuseMask) == 0, "Setting and modulating the same component" );

	if( mSetDiffuseMask != 0 )
	{
		if( mSetDiffuseMask == COL32_MASK_RGBA )
		{
			for(u32 v = 0; v < num_verts; v++)
			{
				p_vertices[v].color = mSetDiffuseColour;
			}
		}
		else
		{
			u32		clear_bits( ~mSetDiffuseMask );
			u32		set_bits( mSetDiffuseColour & mSetDiffuseMask );

			for(u32 v = 0; v < num_verts; v++)
			{
				p_vertices[v].color = (p_vertices[v].color & clear_bits) | set_bits;
			}
		}
	}

	if(mModulateDiffuseMask != 0)
	{
		if( mModulateDiffuseMask == COL32_MASK_RGB )
		{
			for(u32 v = 0; v < num_verts; v++)
			{
				p_vertices[v].color = MODULATE_COLOUR_RGB(p_vertices[v].color, mModulateDiffuseColour);
			}
		}
		else if( mModulateDiffuseMask == COL32_MASK_A )
		{
			for(u32 v = 0; v < num_verts; v++)
			{
				p_vertices[v].color = MODULATE_COLOUR_A(p_vertices[v].color, mModulateDiffuseColour);
			}
		}
		else if( mModulateDiffuseMask == COL32_MASK_RGBA )
		{
			for(u32 v = 0; v < num_verts; v++)
			{
				p_vertices[v].color = MODULATE_COLOUR_RGBA(p_vertices[v].color, mModulateDiffuseColour);
			}
		}
		else
		{
			DAEDALUS_ASSERTMSG( "Unhandled mModulateDiffuseMask" );
		}
	}

}

//*****************************************************************************
//
//*****************************************************************************
void	PSPRenderer::InitCycle1()
{
	DL_PF("CycleType: 1");
	
	BodgeBlendMode( 1 );
}

//*************************************************************************************
// 
//*************************************************************************************
void	PSPRenderer::InitCycle2()
{
	DL_PF("CycleType: 2");

	BodgeBlendMode( 2 );
}

//*****************************************************************************
// States set/modified by this mode:
// SetTextureStateState COLOROP/ALPHAOP
// SetTexture
// D3DRS_TEXTUREFACTOR
// D3DRS_ZENABLE
// D3DRS_ZWRITEENABLE
// ---D3DRS_TEXTUREPERSPECTIVE 
// D3DRS_ALPHATESTENABLE
// D3DRS_SRCBLEND
// D3DRS_DESTBLEND
// D3DRS_ALPHABLENDENABLE
//*****************************************************************************
RDP_OtherMode	gLastRDPOtherMode;
RDP_Mux			gLastRDPMux;
bool			gTextureChanged = false;
u32				gLastPrimColour = 0x00000000;
u32				gLastEnvColour = 0x00000000;

bool			gLastUseZBuffer = false;
bool			gNeedZBufferUdate = false;


void PSPRenderer::InitBlendMode()
{
	DAEDALUS_PROFILE( "PSPRenderer::InitBlendMode" );

	if ( gNeedZBufferUdate ||
		(gRDPOtherMode._u64 != gLastRDPOtherMode._u64) ||
		 (m_bZBuffer != gLastUseZBuffer) )
	{
		// Only update if ZBuffer is enabled
		if (m_bZBuffer)
		{
			if(gRDPOtherMode.z_cmp)
				sceGuEnable(GU_DEPTH_TEST);
			else
				sceGuDisable(GU_DEPTH_TEST);

			sceGuDepthMask( gRDPOtherMode.z_upd ? GL_FALSE : GL_TRUE );	// GL_TRUE to disable z-writes
		}
		else
		{
			sceGuDisable(GU_DEPTH_TEST);
			sceGuDepthMask( GL_TRUE );	// GL_TRUE to disable z-writes
		}

		gNeedZBufferUdate = false;
		gLastUseZBuffer = m_bZBuffer;
	}


	/*if ( gRDPOtherMode._u64 == gLastRDPOtherMode._u64 &&
		 gRDPMux._u64 == gLastRDPMux._u64 &&
		 (gLastPrimColour == m_dwPrimitiveColor ) &&
		 (gLastEnvColour == m_dwEnvColor ) &&
		 gTextureChanged == false )*/
	{
		//static u32 num_ignored = 0;
		//num_ignored++;
		//if ( (num_ignored%10000) == 0)
		//	DBGConsole_Msg( 0, "Ignoring redundant state initialisation %x", num_ignored );
		//return;
	}

	gLastRDPOtherMode._u64 = gRDPOtherMode._u64;
	gLastRDPMux._u64 = gRDPMux._u64;
	gTextureChanged = false;
	gLastPrimColour = m_dwPrimitiveColor;
	gLastEnvColour = m_dwEnvColor;

	switch ( gRDPOtherMode.cycle_type )
	{
		case CYCLE_COPY:			InitCycleCopy();	break;
		case CYCLE_FILL:			InitCycleFill();	break;
		case CYCLE_1CYCLE:			InitCycle1();		break;
		case CYCLE_2CYCLE:			InitCycle2();		break;
	}

	sceGuShadeModel( mSmooth ? GU_SMOOTH : GU_FLAT );
	
	switch ( gRDPOtherMode.text_filt )
	{
		case 0:			//G_TF_POINT:	// 0
			sceGuTexFilter(GU_NEAREST,GU_NEAREST);
			break;
		case 2:			//G_TF_BILERP:	// 2
			sceGuTexFilter(GU_LINEAR,GU_LINEAR);
			break;
		case 3:			//G_TF_AVERAGE:	// 3?
			sceGuTexFilter(GU_LINEAR,GU_LINEAR);
			break;
	}

#define BLEND_NOOP1				0x00000000
#define BLEND_NOOP2				0x00000000

#define BLEND_FOG_ASHADE1		0xc8000000
#define BLEND_FOG_APRIM1		0xc4000000

#define BLEND_PASS1				0x0c080000
#define BLEND_PASS2				0x03020000
	
#define BLEND_OPA1				0x00440000
#define BLEND_OPA2				0x00110000

#define BLEND_XLU1				0x00400000
#define BLEND_XLU2				0x00100000

#define BLEND_ADD1				0x04400000
#define BLEND_ADD2				0x01100000

#define BLEND_MEM1				0x4c400000		// Mem*0 + Mem*(1-0)?!
#define BLEND_MEM2				0x13100000		// Mem*0 + Mem*(1-0)?!


/*

#define	G_BL_CLR_IN	0
#define	G_BL_CLR_MEM	1
#define	G_BL_CLR_BL	2
#define	G_BL_CLR_FOG	3
#define	G_BL_1MA	0
#define	G_BL_A_MEM	1
#define	G_BL_A_IN	0
#define	G_BL_A_FOG	1
#define	G_BL_A_SHADE	2
#define	G_BL_1		2
#define	G_BL_0		3

#define	GBL_c1(m1a, m1b, m2a, m2b)	\
	(m1a) << 30 | (m1b) << 26 | (m2a) << 22 | (m2b) << 18
#define	GBL_c2(m1a, m1b, m2a, m2b)	\
	(m1a) << 28 | (m1b) << 24 | (m2a) << 20 | (m2b) << 16

#define	RM_AA_ZB_OPA_SURF(clk)					\
	AA_EN | Z_CMP | Z_UPD | IM_RD | CVG_DST_CLAMP |		\
	ZMODE_OPA | ALPHA_CVG_SEL |				\
	GBL_c##clk(G_BL_CLR_IN, G_BL_A_IN, G_BL_CLR_MEM, G_BL_A_MEM)

#define	G_RM_FOG_SHADE_A		GBL_c1(G_BL_CLR_FOG, G_BL_A_SHADE, G_BL_CLR_IN, G_BL_1MA)
#define	G_RM_FOG_PRIM_A			GBL_c1(G_BL_CLR_FOG, G_BL_A_FOG, G_BL_CLR_IN, G_BL_1MA)
#define	G_RM_PASS				GBL_c1(G_BL_CLR_IN, G_BL_0, G_BL_CLR_IN, G_BL_1)
#define	RM_AA_ZB_OPA_SURF(clk)	GBL_c##clk(G_BL_CLR_IN, G_BL_A_IN, G_BL_CLR_MEM, G_BL_A_MEM)
*/
	u32 blendmode_1 = u32( gRDPOtherMode._u32[0] & 0xcccc0000 );
	u32 blendmode_2 = u32( gRDPOtherMode._u32[0] & 0x33330000 );

	if ( gRDPOtherMode.cycle_type == CYCLE_FILL )
	{
		sceGuDisable(GU_BLEND);
		sceGuDisable(GU_ALPHA_TEST);
	}
	else if ( blendmode_1 == 0 && blendmode_2 == 0 )
	{
		sceGuBlendFunc(GU_ADD, GU_SRC_ALPHA, GU_ONE_MINUS_SRC_ALPHA, 0, 0);
		sceGuEnable(GU_BLEND);
		sceGuEnable(GU_ALPHA_TEST);
		DL_PF( "      Blend: NOOP/NOOP" );
	}
	else if ( blendmode_1 == BLEND_FOG_ASHADE1 && blendmode_2 == BLEND_OPA2 )
	{
		// c811
		sceGuDisable(GU_BLEND);
		sceGuDisable(GU_ALPHA_TEST);
		DL_PF( "      Blend: FOG_ASHADE/OPA" );
	}
	else if ( blendmode_1 == BLEND_FOG_ASHADE1 && blendmode_2 == BLEND_XLU2 )
	{
		// c810
		sceGuBlendFunc(GU_ADD, GU_SRC_ALPHA, GU_ONE_MINUS_SRC_ALPHA, 0, 0);
		sceGuEnable(GU_BLEND);
		sceGuEnable(GU_ALPHA_TEST);
		DL_PF( "      Blend: FOG_ASHADE/XLU" );
	}
	else if ( blendmode_1 == BLEND_XLU1 && blendmode_2 == BLEND_XLU2 )
	{
		// 0x0050
		sceGuBlendFunc(GU_ADD, GU_SRC_ALPHA, GU_ONE_MINUS_SRC_ALPHA, 0, 0);
		sceGuEnable(GU_BLEND);
		sceGuEnable(GU_ALPHA_TEST);
		DL_PF( "      Blend: XLU/XLU" );
	}
	else if ( blendmode_1 == BLEND_XLU1 && blendmode_2 == BLEND_NOOP2 )
	{
		// 0x0040
		sceGuBlendFunc(GU_ADD, GU_SRC_ALPHA, GU_ONE_MINUS_SRC_ALPHA, 0, 0);
		sceGuEnable(GU_BLEND);
		sceGuEnable(GU_ALPHA_TEST);
		DL_PF( "      Blend: XLU/NOOP" );
	}
	else if ( blendmode_1 == BLEND_XLU1 && blendmode_2 == BLEND_ADD2 )
	{
		// 0x005n
		// XXXX
		sceGuBlendFunc(GU_ADD, GU_SRC_COLOR, GU_DST_COLOR, 0, 0);
		sceGuEnable(GU_BLEND);
		sceGuEnable(GU_ALPHA_TEST);
		DL_PF( "      Blend: XLU/ADD" );
	}
	else if ( blendmode_1 == BLEND_OPA1 && blendmode_2 == BLEND_OPA2 )
	{
		// 0x0055
		// Would have thoughts this was SRCALPHA/DSTALPHA, but apparently not
		sceGuDisable(GU_BLEND);		
		sceGuDisable(GU_ALPHA_TEST);
		DL_PF( "      Blend: OPA/OPA" );
	}
	else if ( blendmode_1 == BLEND_OPA1 && blendmode_2 == BLEND_NOOP2 )
	{
		// 0x0044
		sceGuDisable(GU_BLEND);
		sceGuDisable(GU_ALPHA_TEST);
		DL_PF( "      Blend: OPA/NOOP" );
	}
	else if ( blendmode_1 == BLEND_PASS1 && blendmode_2 == BLEND_PASS2 )
	{
		// 0x0f0a
		sceGuDisable(GU_BLEND);
		sceGuDisable(GU_ALPHA_TEST);
		DL_PF( "      Blend: PASS/PASS" );
	}
	else if ( blendmode_1 == BLEND_PASS1 && blendmode_2 == BLEND_XLU2 )
	{
		// 0x0c18
		sceGuBlendFunc(GU_ADD, GU_SRC_ALPHA, GU_ONE_MINUS_SRC_ALPHA, 0, 0);
		sceGuEnable(GU_BLEND);
		sceGuEnable(GU_ALPHA_TEST);
		DL_PF( "      Blend: PASS/XLU" );
	}
	else if ( blendmode_1 == BLEND_PASS1 && blendmode_2 == BLEND_OPA2 )
	{
		// 0x0c19
		sceGuDisable(GU_BLEND);
		sceGuDisable(GU_ALPHA_TEST);
		DL_PF( "      Blend: PASS/OPA" );
	}
	else if ( blendmode_1 == BLEND_MEM1 && blendmode_2 == BLEND_MEM2 )
	{
		// 0x5f50
		sceGuDisable(GU_BLEND);
		sceGuDisable(GU_ALPHA_TEST);
		DL_PF( "      Blend: MEM/MEM" );
	}
	else
	{
		sceGuBlendFunc(GU_ADD, GU_SRC_ALPHA, GU_ONE_MINUS_SRC_ALPHA, 0, 0);
		sceGuEnable(GU_BLEND);
		sceGuEnable(GU_ALPHA_TEST);
		DL_PF( "      Blend: SRCALPHA/INVSRCALPHA (default: 0x%04x)", gRDPOtherMode.blender );
	}
	

	//
	// I can't think why the hand in mario's menu screen is rendered with an opaque rendermode,
	// and no alpha threshold. We set the alpha reference to 1 to ensure that the transparent pixels
	// don't get rendered. I hope this doesn't fuck anything else up.
	//
	if ( gRDPOtherMode.alpha_compare == 0 )
	{
		if ( gRDPOtherMode.cvg_x_alpha )
		{
			sceGuAlphaFunc(GU_GEQUAL, 0x01, 0xff );
			sceGuEnable(GU_ALPHA_TEST);
		}
		else
		{
			sceGuDisable(GU_ALPHA_TEST);
		}
	}
	else
	{
		// G_AC_THRESHOLD || G_AC_DITHER
		sceGuAlphaFunc(GU_GEQUAL, m_dwAlpha, 0xff );
		sceGuEnable(GU_ALPHA_TEST);
	}

}



//*****************************************************************************
//
//*****************************************************************************
inline f32 round( f32 x )
{
	return (f32)(s32)( x + 0.5f );
}

bool PSPRenderer::TexRect( const v2 & xy0, const v2 & xy1, const v2 & uv0, const v2 & uv1 )
{
	v2 screen0;
	v2 screen1;
	v2 tex_uv0( uv0 );
	v2 tex_uv1( uv1 );

	// Scale to Actual texture coords
	// The two cases are to handle the oversized textures hack on voodoos
#ifdef TEXRECT_USE_3D_TRANSFORM
	float width_scale = (m_fTileWidth[0] > m_fTexWidth[0]) ? m_fInvTileWidth[0] : m_fInvTexWidth[0];
	float height_scale = (m_fTileHeight[0] > m_fTexHeight[0]) ? m_fInvTileHeight[0] : m_fInvTexHeight[0];

	tex_uv0.x *= width_scale;
	tex_uv1.x *= width_scale;

	tex_uv0.y *= height_scale;
	tex_uv1.y *= height_scale;
#endif

	//
	// In Fill/Copy mode the coordinates are inclusive (i.e. add 1.0f to the w/h)
	//
	v2 edge( 0, 0 );

	switch ( gRDPOtherMode.cycle_type )
	{
		case CYCLE_COPY:			edge.x += 1.0f; edge.y += 1.0f;	break;
		case CYCLE_FILL:			edge.x += 1.0f; edge.y += 1.0f;	break;
		case CYCLE_1CYCLE:											break;
		case CYCLE_2CYCLE:											break;
	}


	//
	// We round these value here, so that when we scale up the coords to our screen
	// coords we don't get any gaps.
	//
	screen0.x = round( xy0.x * m_fScreenMultX );
	screen0.y = round( xy0.y * m_fScreenMultY );

	screen1.x = round( (xy1.x + edge.x) * m_fScreenMultX );
	screen1.y = round( (xy1.y + edge.x) * m_fScreenMultY );

#ifdef DEBUG_TEXRECT
	DL_PF( "      ScreenMult:  %f,%f", m_fScreenMultX, m_fScreenMultY );
	DL_PF( "      Screen:  %f,%f -> %f,%f", screen0.x, screen0.y, screen1.x, screen1.y );
	DL_PF( "      Texture: %f,%f -> %f,%f", tex_uv0.x, tex_uv0.y, tex_uv1.x, tex_uv1.y );
#endif

#ifdef TEXRECT_USE_3D_TRANSFORM
	DaedalusVtx trv[4];
	static const u16 indices[2*3] = {1,0,2, 2,0,3};

	trv[0].pos = v3( screen0.x, screen0.y, gTexRectDepth );
	trv[0].color = 0xffffffff;
	trv[0].t0 = v2( tex_uv0.x, tex_uv0.y );

	trv[1].pos = v3( screen1.x, screen0.y, gTexRectDepth );
	trv[1].color = 0xffffffff;
	trv[1].t0 = v2( tex_uv1.x, tex_uv0.y );

	trv[2].pos = v3( screen1.x, screen1.y, gTexRectDepth );
	trv[2].color = 0xffffffff;
	trv[2].t0 = v2( tex_uv1.x, tex_uv1.y );

	trv[3].pos = v3( screen0.x, screen1.y, gTexRectDepth );
	trv[3].color = 0xffffffff;
	trv[3].t0 = v2( tex_uv0.x, tex_uv1.y );

	return RenderTriangleList( trv, 4, indices, 6, true, true );
#else
	DaedalusVtx trv[2];

	trv[0].pos = v3( screen0.x, screen0.y, gTexRectDepth );
	trv[0].color = 0xffffffff;
	trv[0].t0 = v2( tex_uv0.x, tex_uv0.y );

	trv[1].pos = v3( screen1.x, screen1.y, gTexRectDepth );
	trv[1].color = 0xffffffff;
	trv[1].t0 = v2( tex_uv1.x, tex_uv1.y );

	return RenderSprites( trv, 2, true, true );
#endif
}

//*****************************************************************************
//
//*****************************************************************************
bool PSPRenderer::TexRectFlip( const v2 & xy0, const v2 & xy1, const v2 & uv0, const v2 & uv1 )
{
	// TODO - In 1/2cycle mode, skip bottom/right edges!?
	v2 screen0;
	v2 screen1;
	v2 tex_uv0( uv0 );
	v2 tex_uv1( uv1 );

	// Scale to Actual texture coords
	// The two cases are to handle the oversized textures hack on voodoos
#ifdef TEXRECT_USE_3D_TRANSFORM
	float width_scale = (m_fTileWidth[0] > m_fTexWidth[0]) ? m_fInvTileWidth[0] : m_fInvTexWidth[0];
	float height_scale = (m_fTileHeight[0] > m_fTexHeight[0]) ? m_fInvTileHeight[0] : m_fInvTexHeight[0];

	tex_uv0.x *= width_scale;
	tex_uv1.x *= width_scale;

	tex_uv0.y *= height_scale;
	tex_uv1.y *= height_scale;
#endif

	//
	// In Fill/Copy mode the coordinates are inclusive (i.e. add 1.0f to the w/h)
	//
	v2 edge( 0, 0 );

	switch ( gRDPOtherMode.cycle_type )
	{
		case CYCLE_COPY:			edge.x += 1.0f; edge.y += 1.0f;	break;
		case CYCLE_FILL:			edge.x += 1.0f; edge.y += 1.0f;	break;
		case CYCLE_1CYCLE:											break;
		case CYCLE_2CYCLE:											break;
	}

	//
	// We round these value here, so that when we scale up the coords to our screen
	// coords we don't get any gaps.
	//
	screen0.x = round( xy0.x * m_fScreenMultX );
	screen0.y = round( xy0.y * m_fScreenMultY );

	screen1.x = round( (xy1.x + edge.x) * m_fScreenMultX );
	screen1.y = round( (xy1.y + edge.x) * m_fScreenMultY );

#ifdef DEBUG_TEXRECT
	DL_PF( "      ScreenMult:  %f,%f", m_fScreenMultX, m_fScreenMultY );
	DL_PF( "      Screen:  %f,%f -> %f,%f", screen0.x, screen0.y, screen1.x, screen1.y );
	DL_PF( "      Texture: %f,%f -> %f,%f", tex_uv0.x, tex_uv0.y, tex_uv1.x, tex_uv1.y );
#endif

#ifdef TEXRECT_USE_3D_TRANSFORM
	DaedalusVtx trv[4];
	static const u16 indices[2*3] = {1,0,2, 2,0,3};

	// Same as TexRect, but with texcoords 0,2 swapped
	trv[0].pos = v3( screen0.x, screen0.y, gTexRectDepth );
	trv[0].color = 0xffffffff;
	trv[0].t0 = v2( tex_uv0.x, tex_uv0.y );
	
	trv[1].pos = v3( screen1.x, screen0.y, gTexRectDepth );
	trv[1].color = 0xffffffff;
	trv[1].t0 = v2( tex_uv0.x, tex_uv1.y );		// In TexRect this is tex_uv1.x, tex_uv0.y

	trv[2].pos = v3( screen1.x, screen1.y, gTexRectDepth );
	trv[2].color = 0xffffffff;
	trv[2].t0 = v2( tex_uv1.x, tex_uv1.y );

	trv[3].pos = v3( screen0.x, screen1.y, gTexRectDepth );
	trv[3].color = 0xffffffff;
	trv[3].t0 = v2( tex_uv1.x, tex_uv0.y );		//tex_uv0.x	tex_uv1.y	

	return RenderTriangleList( trv, 4, indices, 6, true, true );
#else
	DaedalusVtx trv[2];

	trv[0].pos = v3( screen0.x, screen0.y, gTexRectDepth );
	trv[0].color = 0xffffffff;
	trv[0].t0 = v2( tex_uv0.x, tex_uv0.y );

	trv[1].pos = v3( screen1.x, screen1.y, gTexRectDepth );
	trv[1].color = 0xffffffff;
	trv[1].t0 = v2( tex_uv1.x, tex_uv1.y );

	return RenderSprites( trv, 2, true, true );
#endif
}

//*****************************************************************************
//
//*****************************************************************************
bool PSPRenderer::FillRect( const v2 & xy0, const v2 & xy1, u32 color )
{
	v2 screen0;
	v2 screen1;

	if ( (gRDPOtherMode._u32[0] & 0xffff0000) == 0x5f500000 )
	{
		// this blend mode is mem*0 + mem*1, so we don't need to render it... Very odd!
		return true;
	}

	// This if for C&C - It might break other stuff (I'm not sure if we should allow alpha or not..)
//	color |= 0xff000000;

	//
	// We round these value here, so that when we scale up the coords to our screen
	// coords we don't get any gaps.
	//
	screen0.x = round( xy0.x * m_fScreenMultX );
	screen0.y = round( xy0.y * m_fScreenMultY );

	screen1.x = round( xy1.x * m_fScreenMultX );
	screen1.y = round( xy1.y * m_fScreenMultY );

	DL_PF( "      ScreenMult:  %f,%f", m_fScreenMultX, m_fScreenMultY );
	DL_PF( "      Screen:  %f,%f -> %f,%f", screen0.x, screen0.y, screen1.x, screen1.y );

#ifdef TEXRECT_USE_3D_TRANSFORM
	DaedalusVtx frv[4];
	static const u16 indices[2*3] = {1,0,2, 2,0,3};

	frv[0].pos = v3(screen0.x, screen0.y, gFillRectDepth );
	frv[0].color = color;
	frv[0].t0 = v2( 0, 0 );

	frv[1].pos = v3(screen1.x, screen0.y, gFillRectDepth );
	frv[1].color = color;
	frv[1].t0 = v2( 1, 0 );

	frv[2].pos = v3(screen1.x, screen1.y, gFillRectDepth );
	frv[2].color = color;
	frv[2].t0 = v2( 1, 1 );

	frv[3].pos = v3(screen0.x, screen1.y, gFillRectDepth );
	frv[3].color = color;
	frv[3].t0 = v2( 0, 1 );

	return RenderTriangleList( frv, 4, indices, 6, true );
#else
	DaedalusVtx trv[2];

	trv[0].pos = v3( screen0.x, screen0.y, gFillRectDepth );
	trv[0].color = color;
	trv[0].t0 = v2( 0, 0 );

	trv[1].pos = v3( screen1.x, screen1.y, gFillRectDepth );
	trv[1].color = color;
	trv[1].t0 = v2( 1, 1 );

	return RenderSprites( trv, 2, true, false );
#endif
}

//*****************************************************************************
// Returns true if triangle visible and rendered, false otherwise
//*****************************************************************************
bool PSPRenderer::AddTri(u32 v0, u32 v1, u32 v2)
{
	if ( TestTri( v0, v1, v2 ) )
	{
		DL_PF("   Tri: %d,%d,%d", v0, v1, v2);

		m_swIndexBuffer[ m_dwNumIndices + 0 ] = (u16)v0;
		m_swIndexBuffer[ m_dwNumIndices + 1 ] = (u16)v1;
		m_swIndexBuffer[ m_dwNumIndices + 2 ] = (u16)v2;

		m_dwNumIndices += 3;

		mVtxClipFlagsUnion |= mVtxClipFlags[v0];
		mVtxClipFlagsUnion |= mVtxClipFlags[v1];
		mVtxClipFlagsUnion |= mVtxClipFlags[v2];

		m_dwNumTrisRendered++;
		return true;
	}
	else
	{
		DL_PF("   Tri: %d,%d,%d (clipped)", v0, v1, v2);

		m_dwNumTrisClipped++;
		return false;
	}
}

//*****************************************************************************
//
//*****************************************************************************
// Returns true if it thinks the triangle is visible
// Returns false if it is clipped
bool PSPRenderer::TestTri(u32 v0, u32 v1, u32 v2) const
{
#if 0
	// Check vertices are valid!
	if ( v0 >= MAX_VERTS || v1 >= MAX_VERTS || v2 >= MAX_VERTS )
		return false;
#endif

	// Here we AND all the flags. If any of the bits is set for all
	// 3 vertices, it means that all three x, y or z lie outside of
	// the current viewing volume.
	if (mVtxClipFlags[v0] & mVtxClipFlags[v1] & mVtxClipFlags[v2] & CLIP_TEST_FLAGS)
	{
		return false;
	}

	// The backface cull is disabled for now. This is likely to be faster in hardware.
	//if ((m_bCullFront || m_bCullBack) && !IsZModeDecal())
	//{
	//	const v4 & vert0 = mVtxProjected[v0].ProjectedPos;
	//	const v4 & vert1 = mVtxProjected[v1].ProjectedPos;
	//	const v4 & vert2 = mVtxProjected[v2].ProjectedPos;

	//	float v0x = vert0.x / vert0.w;
	//	float v0y = vert0.y / vert0.w;
	//	float v1x = vert1.x / vert1.w;
	//	float v1y = vert1.y / vert1.w;
	//	float v2x = vert2.x / vert2.w;
	//	float v2y = vert2.y / vert2.w;

	//	float V1 = v2x - v0x;
	//	float V2 = v2y - v0y;
	//	
	//	float W1 = v2x - v1x;
	//	float W2 = v2y - v1y;

	//	//
	//	//	Sign flipped to avoid extra float op
	//	//
	//	float fDirection = ( (V1 * W2) - (V2 * W1) );

	//	// Account for being behind the camera
	//	fDirection *= vert0.w * vert1.w * vert2.w;

	//	if (fDirection < 0 && m_bCullBack)
	//		return false;
	//	else if (fDirection > 0 && m_bCullFront)
	//		return false;
	//}
	return true;
}


//*****************************************************************************
//
//*****************************************************************************
bool PSPRenderer::TestVerts( u32 v0, u32 vn ) const
{
	u32 flags = mVtxClipFlags[v0];

	for ( u32 i = v0+1; i <= vn && i < MAX_VERTS; i++ )
	{
		flags &= mVtxClipFlags[i];
	}

	return (flags & CLIP_TEST_FLAGS) == 0;
}

//*****************************************************************************
//
//*****************************************************************************
v4 PSPRenderer::LightVert( const v3 & norm ) const
{
	// Do ambient
	v4	result( mAmbientColour );

	for ( s32 l = 0; l < m_dwNumLights; l++ )
	{
		f32 fCosT = norm.Dot( mLights[l].Direction );
		if (fCosT > 0)
		{
			result.x += mLights[l].Colour.x * fCosT;
			result.y += mLights[l].Colour.y * fCosT;
			result.z += mLights[l].Colour.z * fCosT;
		//	result.w += mLights[l].Colour.w * fCosT;
		}
	}

	if( result.x > 1.0f ) result.x = 1.0f;
	if( result.y > 1.0f ) result.y = 1.0f;
	if( result.z > 1.0f ) result.z = 1.0f;
	if( result.w > 1.0f ) result.w = 1.0f;

	return result;
}

//*****************************************************************************
//
//	The following clipping code was taken from The Irrlicht Engine.
//	See http://irrlicht.sourceforge.net/ for more information.
//	Copyright (C) 2002-2006 Nikolaus Gebhardt/Alten Thomas
//
//*****************************************************************************
static const v4 NDCPlane[6] =
{
	v4(  0.f,  0.f,  1.f, -1.f ),	// near
	v4(  1.f,  0.f,  0.f, -1.f ),	// left
	v4( -1.f,  0.f,  0.f, -1.f ),	// right
	v4(  0.f,  1.f,  0.f, -1.f ),	// bottom
	v4(  0.f, -1.f,  0.f, -1.f ),	// top
	v4(  0.f,  0.f, -1.f, -1.f )	// far
};

static u32 clipToHyperPlane( DaedalusVtx4 * dest, const DaedalusVtx4 * source, u32 inCount, const v4 &plane )
{
	u32 outCount;
	DaedalusVtx4 * out;

	const DaedalusVtx4 * a;
	const DaedalusVtx4 * b;

	f32 aDotPlane;
	f32 bDotPlane;

	out = dest;
	outCount = 0;
	b = source;
	bDotPlane = b->ProjectedPos.Dot( plane );

	for( u32 i = 1; i < inCount + 1; ++i)
	{
		a = &source[i%inCount];
		aDotPlane = a->ProjectedPos.Dot( plane );

		// current point inside
		if ( aDotPlane <= 0.f )
		{
			// last point outside
			if ( bDotPlane > 0.f )
			{
				// intersect line segment with plane
				out->Interpolate( *b, *a, bDotPlane / (b->ProjectedPos - a->ProjectedPos).Dot( plane ) );
				out += 1;
				outCount += 1;
			}
			// copy current to out
			*out = *a;
			b = out;

			out += 1;
			outCount += 1;
		}
		else
		{
			// current point outside

			if ( bDotPlane <= 0.f )
			{
				// previous was inside
				// intersect line segment with plane
				out->Interpolate( *b, *a, bDotPlane / (b->ProjectedPos - a->ProjectedPos).Dot( plane ) );

				out += 1;
				outCount += 1;
			}
			b = a;
		}

		bDotPlane = b->ProjectedPos.Dot( plane );
	}

	return outCount;
}

u32 clip_to_frustum( DaedalusVtx4 * v0, DaedalusVtx4 * v1, u32 vIn )
{
	u32 vOut;

	vOut = vIn;

	vOut = clipToHyperPlane( v1, v0, vOut, NDCPlane[2] );		// right
	vOut = clipToHyperPlane( v0, v1, vOut, NDCPlane[1] );		// left
	vOut = clipToHyperPlane( v1, v0, vOut, NDCPlane[4] );		// top
	vOut = clipToHyperPlane( v0, v1, vOut, NDCPlane[3] );		// bottom
	vOut = clipToHyperPlane( v1, v0, vOut, NDCPlane[0] );		// near
	vOut = clipToHyperPlane( v0, v1, vOut, NDCPlane[5] );		// far

	return vOut;
}

u32 clip_to_frustum_vfpu( DaedalusVtx4 * v0, DaedalusVtx4 * v1, u32 vIn )
{
	u32 vOut;

	vOut = vIn;

	vOut = _ClipToHyperPlane( v1, v0, &NDCPlane[2], vOut );		// right
	vOut = _ClipToHyperPlane( v0, v1, &NDCPlane[1], vOut );		// left
	vOut = _ClipToHyperPlane( v1, v0, &NDCPlane[4], vOut );		// top
	vOut = _ClipToHyperPlane( v0, v1, &NDCPlane[3], vOut );		// bottom
	vOut = _ClipToHyperPlane( v1, v0, &NDCPlane[0], vOut );		// near
	vOut = _ClipToHyperPlane( v0, v1, &NDCPlane[5], vOut );		// far

	return vOut;
}
//*****************************************************************************
//
//*****************************************************************************
bool PSPRenderer::FlushTris()
{
	if ( m_dwNumIndices == 0 )
	{
		mVtxClipFlagsUnion = 0;
		return true;
	}

	u32				num_vertices;
	DaedalusVtx *	p_vertices;

	if(gSoftwareClipping && (mVtxClipFlagsUnion != 0))
	{
		PrepareTrisClipped( &p_vertices, &num_vertices );
	}
	else
	{
		PrepareTrisUnclipped( &p_vertices, &num_vertices );
	}

	//
	// Process the software vertex buffer to apply a couple of 
	// necessary changes to the texture coords (this is required
	// because some ucodes set the texture after setting the vertices)
	//
	if (m_bTextureEnable)
	{
		// Only scale if not generated automatically
		if ( !( m_bTextureGen && m_bLightingEnable ) )
		{
			for ( u32 v = 0; v < num_vertices; v++ )
			{
				p_vertices[v].t0.x = (p_vertices[v].t0.x - m_fTexLeft[0]) * m_fInvTexWidth[0];
				p_vertices[v].t0.y = (p_vertices[v].t0.y - m_fTexTop[0]) * m_fInvTexHeight[0];
			}
		}

		// Bias points in decal mode
		if (IsZModeDecal())
		{
			for ( u32 v = 0; v < num_vertices; v++ )
			{
				p_vertices[v].pos.z += DECAL_Z_OFFSET;
			}
		}
	}

	//
	// If smooth shading is turned off, duplicate the first diffuse color for all three verts
	//
	if ( !mSmoothShade )
	{

	}

	{
		DAEDALUS_PROFILE( "Prepare Blend" );

		InitBlendMode();
		PrepareVerticesForBlendMode( p_vertices, num_vertices );
	}

	//
	//	For now do all clipping though ge - 
	//
	if(m_bCullFront)
	{
		sceGuFrontFace(GU_CW);
		sceGuEnable(GU_CULL_FACE);
	}
	else if( m_bCullBack )
	{
		sceGuFrontFace(GU_CCW);
		sceGuEnable(GU_CULL_FACE);
	}
	else
	{
		sceGuDisable(GU_CULL_FACE);
	}

	//
	//	Render out our vertices
	//
	const Matrix4x4	&		projection( m_mProjection[m_dwProjectionTop] );
	const ScePspFMatrix4 *	p_psp_proj( reinterpret_cast< const ScePspFMatrix4 * >( &projection ) );

	sceGumMatrixMode(GU_PROJECTION);
	sceGumLoadMatrix( p_psp_proj );

	sceGumDrawArray(GU_TRIANGLES,DAEDALUS_VERTEX_FLAGS|GU_TRANSFORM_3D,num_vertices,NULL,p_vertices);
	sceGuDisable(GU_CULL_FACE);

	m_dwNumIndices = 0;
	mVtxClipFlagsUnion = 0;
	return true;
}

//*****************************************************************************
//
//*****************************************************************************
namespace	
{
	DaedalusVtx4		temp_a[ 10 ];
	DaedalusVtx4		temp_b[ 10 ];
}

//*****************************************************************************
//
//*****************************************************************************
void PSPRenderer::PrepareTrisClipped( DaedalusVtx ** p_p_vertices, u32 * p_num_vertices ) const
{
	DAEDALUS_PROFILE( "PSPRenderer::PrepareTrisClipped" );

	pspvfpu_use_matrices(gDaedalusVFPUContext, 0, VMAT0 | VMAT1 | VMAT2 | VMAT3 );

	//
	//	At this point all vertices are lit/projected and have both transformed and projected
	//	vertex positions. For the best results we clip against the projected vertex positions,
	//	but use the resulting intersections to interpolate the transformed positions. 
	//	The clipping is more efficient in normalised device coordinates, but rendering these
	//	directly prevents the PSP performing perspective correction. We could invert the projection
	//	matrix and use this to back-project the clip planes into world coordinates, but this
	//	suffers from various precision issues. Carrying around both sets of coordinates gives
	//	us the best of both worlds :)
	//
	static std::vector<DaedalusVtx4>		clipped_vertices;
	clipped_vertices.erase( clipped_vertices.begin(), clipped_vertices.end() );

	for(u32 i = 0; i < m_dwNumIndices; i+=3)
	{
		u32 idx0 = m_swIndexBuffer[ i + 0 ];
		u32 idx1 = m_swIndexBuffer[ i + 1 ];
		u32 idx2 = m_swIndexBuffer[ i + 2 ];

		if(mVtxClipFlags[idx0] | mVtxClipFlags[idx1] | mVtxClipFlags[idx2])
		{
			temp_a[ 0 ] = mVtxProjected[ idx0 ];
			temp_a[ 1 ] = mVtxProjected[ idx1 ];
			temp_a[ 2 ] = mVtxProjected[ idx2 ];

			DL_PF( "i%d: %f,%f,%f,%f", i+0, temp_a[0].ProjectedPos.x, temp_a[0].ProjectedPos.y, temp_a[0].ProjectedPos.z, temp_a[0].ProjectedPos.w );
			DL_PF( "i%d: %f,%f,%f,%f", i+1, temp_a[1].ProjectedPos.x, temp_a[1].ProjectedPos.y, temp_a[1].ProjectedPos.z, temp_a[1].ProjectedPos.w );
			DL_PF( "i%d: %f,%f,%f,%f", i+2, temp_a[2].ProjectedPos.x, temp_a[2].ProjectedPos.y, temp_a[2].ProjectedPos.z, temp_a[2].ProjectedPos.w );

			//u32 out = clip_to_frustum( temp_a, temp_b, 3 );
			u32 out = clip_to_frustum_vfpu( temp_a, temp_b, 3 );
			if( out < 3 )
				continue;

			// Retesselate
			for( u32 j = 0; j <= out - 3; ++j )
			{
				clipped_vertices.push_back( temp_a[ 0 ] );
				clipped_vertices.push_back( temp_a[ j + 1 ] );
				clipped_vertices.push_back( temp_a[ j + 2 ] );
			}
		}
		else
		{
			clipped_vertices.push_back( mVtxProjected[ idx0 ] );
			clipped_vertices.push_back( mVtxProjected[ idx1 ] );
			clipped_vertices.push_back( mVtxProjected[ idx2 ] );
		}
	}

	//
	//	Now the vertices have been clipped we need to write them into
	//	a buffer we obtain this from the display list.
	//	Maybe we should allocate all vertex buffers from VRAM?
	//
	u32				num_vertices( clipped_vertices.size() );
	DaedalusVtx *	p_vertices( (DaedalusVtx*)sceGuGetMemory(num_vertices*sizeof(DaedalusVtx)) );
	for( u32 i = 0; i < num_vertices; ++i )
	{
		p_vertices[ i ].t0 = clipped_vertices[ i ].t0;
		p_vertices[ i ].color = ColourU32::Vector2ColourClamped( clipped_vertices[ i ].color );
		p_vertices[ i ].pos.x = clipped_vertices[ i ].TransformedPos.x;
		p_vertices[ i ].pos.y = clipped_vertices[ i ].TransformedPos.y;
		p_vertices[ i ].pos.z = clipped_vertices[ i ].TransformedPos.z;
	}

	*p_p_vertices = p_vertices;
	*p_num_vertices = num_vertices;
}

//*****************************************************************************
//
//*****************************************************************************
void PSPRenderer::PrepareTrisUnclipped( DaedalusVtx ** p_p_vertices, u32 * p_num_vertices ) const
{
	DAEDALUS_PROFILE( "PSPRenderer::PrepareTrisUnclipped" );
	DAEDALUS_ASSERT( m_dwNumIndices > 0, "The number of indices should have been checked" );

	u32				num_vertices( m_dwNumIndices );
	DaedalusVtx *	p_vertices( (DaedalusVtx*)sceGuGetMemory(num_vertices*sizeof(DaedalusVtx)) );

	//
	//	Previously this code set up an index buffer to avoid processing the
	//	same vertices more than once - we avoid this now as there is apparently
	//	quite a large performance penalty associated with using these on the PSP.
	//
	//	http://forums.ps2dev.org/viewtopic.php?t=4703
	//

	//
	//	Now we just shuffle all the data across directly (potentially duplicating verts)
	//
	for( u32 i = 0; i < m_dwNumIndices; ++i )
	{
		u32			index( m_swIndexBuffer[ i ] );

		p_vertices[ i ].t0 = mVtxProjected[ index ].t0;
		p_vertices[ i ].color = ColourU32::Vector2ColourClamped( mVtxProjected[ index ].color );
		p_vertices[ i ].pos.x = mVtxProjected[ index ].TransformedPos.x;
		p_vertices[ i ].pos.y = mVtxProjected[ index ].TransformedPos.y;
		p_vertices[ i ].pos.z = mVtxProjected[ index ].TransformedPos.z;
	}

	*p_p_vertices = p_vertices;
	*p_num_vertices = num_vertices;
}

//*****************************************************************************
// Used for TexRect, TexRectFlip, FillRect
//*****************************************************************************
bool	PSPRenderer::RenderTriangleList( const DaedalusVtx * p_verts, u32 num_verts, const u16 * p_indices, u32 num_indices, bool disable_zbuffer, bool force_point_filter )
{
	if ( disable_zbuffer )
	{
		gNeedZBufferUdate = true;
	}

	if ( force_point_filter )
	{
		// Mariokart hack to make bg on main screen work well Can do this or just clamp values
		// Need filter update!
	}
	

	DaedalusVtx*	p_vertices = (DaedalusVtx*)sceGuGetMemory(num_verts*sizeof(DaedalusVtx));
	u16 *			p_sw_indices = (u16*)sceGuGetMemory(num_indices*sizeof(u16));

	memcpy( p_vertices, p_verts, num_verts*sizeof(DaedalusVtx));
	memcpy( p_sw_indices, p_indices, num_indices*sizeof(u16));

	InitBlendMode();
	PrepareVerticesForBlendMode( p_vertices, num_verts );

	if ( disable_zbuffer )
	{
		sceGuDisable(GU_DEPTH_TEST);
		sceGuDepthMask( GL_TRUE );	// GL_TRUE to disable z-writes
	}

	if ( force_point_filter )
	{
		// Mariokart hack to make bg on main screen work well Can do this or just clamp values
		sceGuTexFilter(GU_NEAREST,GU_NEAREST);
	}


	//sceGuEnable(GU_ALPHA_TEST);
	//sceGuBlendFunc(GU_ADD, GU_SRC_ALPHA, GU_ONE_MINUS_SRC_ALPHA, 0, 0);
	//sceGuEnable(GU_BLEND);
	//sceGuTexFunc(GU_TFX_MODULATE,GU_TCC_RGB);
	//sceGuTexFunc(GU_TFX_REPLACE,GU_TCC_RGB);
	//sceGuTexImage(0,gPlaceholderTextureWidth,gPlaceholderTextureHeight,gPlaceholderTextureWidth,gWhiteTexture);

	//sceGuDisable(GU_BLEND);
	//sceGuDisable(GU_DEPTH_TEST);

	sceGumMatrixMode(GU_PROJECTION);
	sceGumLoadIdentity();
	sceGumOrtho(0, 480, 272, 0, 0.0f, -1.0f);

	sceGumDrawArray(GU_TRIANGLES,DAEDALUS_VERTEX_FLAGS|GU_TRANSFORM_3D|GU_INDEX_16BIT,num_indices,p_sw_indices,p_vertices);

	return true;
}

//*****************************************************************************
// Used for TexRect, TexRectFlip, FillRect
//*****************************************************************************
bool	PSPRenderer::RenderSprites( const DaedalusVtx * p_verts, u32 num_verts, bool disable_zbuffer, bool force_point_filter )
{
	if ( disable_zbuffer )
	{
		gNeedZBufferUdate = true;
	}

	if ( force_point_filter )
	{
		// Mariokart hack to make bg on main screen work well Can do this or just clamp values
		// Need filter update!
	}
	
	DaedalusVtx*	p_vertices = (DaedalusVtx*)sceGuGetMemory(num_verts*sizeof(DaedalusVtx));
	memcpy( p_vertices, p_verts, num_verts*sizeof(DaedalusVtx));

	InitBlendMode();
	PrepareVerticesForBlendMode( p_vertices, num_verts );

	if ( disable_zbuffer )
	{
		sceGuDisable(GU_DEPTH_TEST);
		sceGuDepthMask( GL_TRUE );	// GL_TRUE to disable z-writes
	}

	if ( force_point_filter )
	{
		// Mariokart hack to make bg on main screen work well Can do this or just clamp values
		sceGuTexFilter(GU_NEAREST,GU_NEAREST);
	}



	//sceGuEnable(GU_ALPHA_TEST);
	//sceGuBlendFunc(GU_ADD, GU_SRC_ALPHA, GU_ONE_MINUS_SRC_ALPHA, 0, 0);
	//sceGuEnable(GU_BLEND);
	//sceGuTexFunc(GU_TFX_MODULATE,GU_TCC_RGB);
	//sceGuTexFunc(GU_TFX_REPLACE,GU_TCC_RGB);
	//sceGuTexImage(0,gPlaceholderTextureWidth,gPlaceholderTextureHeight,gPlaceholderTextureWidth,gWhiteTexture);

	//sceGuDisable(GU_BLEND);
	//sceGuDisable(GU_DEPTH_TEST);

	sceGumMatrixMode(GU_PROJECTION);
	sceGumLoadIdentity();
	//sceGumOrtho(0, 480 +1, 272 +1, 0, 0.0f, -1.0f);

	//sceGumDrawArray(GU_TRIANGLES,DAEDALUS_VERTEX_FLAGS|GU_TRANSFORM_3D|GU_INDEX_16BIT,num_indices,p_sw_indices,p_vertices);
	sceGumDrawArray(GU_SPRITES,DAEDALUS_VERTEX_FLAGS|GU_TRANSFORM_2D,num_verts,NULL,p_vertices);

	return true;
}

//*****************************************************************************
//
//*****************************************************************************
void PSPRenderer::SetNewVertexInfo(u32 dwAddress, u32 dwV0, u32 dwNum)
{
	if(gUseVFPUTnL)
	{
		SetNewVertexInfoVFPU(dwAddress, dwV0, dwNum);
	}
	else
	{
		SetNewVertexInfoCPU(dwAddress, dwV0, dwNum);
	}
}

//*****************************************************************************
//
//*****************************************************************************
void PSPRenderer::SetNewVertexInfoVFPU(u32 dwAddress, u32 dwV0, u32 dwNum)
{
	pspvfpu_use_matrices(gDaedalusVFPUContext, 0, VMAT0 | VMAT1 | VMAT2 | VMAT3 );

	//DBGConsole_Msg(0, "In SetNewVertexInfo");
	const FiddledVtx * const pVtxBase( (const FiddledVtx*)(g_pu8RamBase + dwAddress) );

	const Matrix4x4 & matWorld( m_mModelView[m_dwModelViewTop] );
	const Matrix4x4 & matProject( m_mProjection[m_dwProjectionTop] );

	if (m_bLightingEnable)
	{
		_TransformVerticesWithLighting( &matWorld, &matProject, pVtxBase, &mVtxProjected[dwV0], dwNum, 
										&mAmbientColour, mLights, m_dwNumLights );
	}
	else
	{
		_TransformVerticesWithColour( &matWorld, &matProject, pVtxBase, &mVtxProjected[dwV0], dwNum );
	}

	if(m_bFogEnable || m_bTextureEnable)
	{
		for (u32 i = dwV0; i < dwV0 + dwNum; i++)
		{
			const FiddledVtx & vert = pVtxBase[i - dwV0];
			const v4 &	projected( mVtxProjected[i].ProjectedPos );

			// I decided to take this out for v0.06b release - it's not quite
			// stable enough yet
			if (m_bFogEnable)
			{
				float	rhw;
				if(fabsf(projected.w) > 0.0f)
				{
					rhw = 1.0f / ( projected.w );
				}
				else
				{
					rhw = 0.0f;
				}

				float eyespace_z = projected.z * rhw;
				float fFog = (eyespace_z * m_fFogMult) + m_fFogOffset;

				f32 fog_coeff = (f32)(fFog);
				f32 alpha = Clamp< f32 >( fog_coeff, 0.0f, 1.0f );

				// Set the alpha
				mVtxProjected[i].color.w = alpha;
			}

			if (m_bTextureEnable)
			{
				// Update texture coords n.b. need to divide tu/tv by bogus scale on addition to buffer

				// If the vert is already lit, then there is no normal (and hence we
				// can't generate tex coord)
				if (m_bTextureGen && m_bLightingEnable)
				{
					v3 vecTransformedNormal;		// Used only when m_bLightingEnable
					v3	model_normal(f32( vert.norm_x ), f32( vert.norm_y ), f32( vert.norm_z ) );

					vecTransformedNormal = matWorld.TransformNormal( model_normal );
					vecTransformedNormal.Normalise();

					const v3 & norm = vecTransformedNormal;

					// Assign the spheremap's texture coordinates
					mVtxProjected[i].t0.x = (0.5f * ( 1.0f + ( norm.x*matWorld.m11 +
															   norm.y*matWorld.m21 +
															   norm.z*matWorld.m31 ) ))/* * m_fTexScaleX*/;

					mVtxProjected[i].t0.y = (0.5f * ( 1.0f - ( norm.x*matWorld.m12 +
															   norm.y*matWorld.m22 +
															   norm.z*matWorld.m32 ) ))/* * m_fTexScaleY*/;
				}
				else
				{
					mVtxProjected[i].t0.x = (float)vert.tu * m_fTexScaleX;
					mVtxProjected[i].t0.y = (float)vert.tv * m_fTexScaleY;
				}

				//mVtxProjected[i].t1 = mVtxProjected[i].t0;
			}
		}
	}

	//
	//	Initialise the clipping flags.
	//
	for (u32 i = dwV0; i < dwV0 + dwNum; i++)
	{
		const v4 &	projected( mVtxProjected[i].ProjectedPos );

		u32 clip_flags = 0;
		if		(-projected.x > projected.w)	clip_flags |= X_POS;
		else if (+projected.x > projected.w)	clip_flags |= X_NEG;

		if		(-projected.y > projected.w)	clip_flags |= Y_POS;
		else if (+projected.y > projected.w)	clip_flags |= Y_NEG;

		if		(-projected.z > projected.w)	clip_flags |= Z_POS;
		else if (+projected.z > projected.w)	clip_flags |= Z_NEG;
		mVtxClipFlags[i] = clip_flags;
	}
}


//*****************************************************************************
//
//*****************************************************************************
void PSPRenderer::SetNewVertexInfoCPU(u32 dwAddress, u32 dwV0, u32 dwNum)
{
	//DBGConsole_Msg(0, "In SetNewVertexInfo");
	FiddledVtx * pVtxBase = (FiddledVtx*)(g_pu8RamBase + dwAddress);

	const Matrix4x4 & matWorld		= m_mModelView[m_dwModelViewTop];
	const Matrix4x4 matWorldProject( m_mModelView[m_dwModelViewTop] * m_mProjection[m_dwProjectionTop] );


	// Transform and Project + Lighting
	// or 
	// Transform and Project with Colour

	u32 i;
	for (i = dwV0; i < dwV0 + dwNum; i++)
	{
		const FiddledVtx & vert = pVtxBase[i - dwV0];

		v4		w( f32( vert.x ), f32( vert.y ), f32( vert.z ), 1.0f );

		mVtxProjected[i].ProjectedPos = matWorldProject.Transform( w );
		mVtxProjected[i].TransformedPos = matWorld.Transform( w );

		const v4 &	projected( mVtxProjected[i].ProjectedPos );

		DL_PF( "p%d: (%f,%f,%f) -> (%f,%f,%f,%f)", i, w.x, w.y, w.z, projected.x, projected.y, projected.z, projected.w );

		u32 clip_flags = 0;
		if		(-projected.x > projected.w)	clip_flags |= X_POS;
		else if (+projected.x > projected.w)	clip_flags |= X_NEG;

		if		(-projected.y > projected.w)	clip_flags |= Y_POS;
		else if (+projected.y > projected.w)	clip_flags |= Y_NEG;

		if		(-projected.z > projected.w)	clip_flags |= Z_POS;
		else if (+projected.z > projected.w)	clip_flags |= Z_NEG;
		mVtxClipFlags[i] = clip_flags;

		v3 vecTransformedNormal;		// Used only when m_bLightingEnable
		if (m_bLightingEnable)
		{
			v3	model_normal(f32( vert.norm_x ), f32( vert.norm_y ), f32( vert.norm_z ) );

			vecTransformedNormal = matWorld.TransformNormal( model_normal );
			vecTransformedNormal.Normalise();

			mVtxProjected[i].color = LightVert(vecTransformedNormal);
		}
		else
		{
			mVtxProjected[i].color = v4( vert.rgba_r / 255.0f, vert.rgba_g / 255.0f, vert.rgba_b / 255.0f, vert.rgba_a / 255.0f );
		}

		// I decided to take this out for v0.06b release - it's not quite
		// stable enough yet
		if (m_bFogEnable)
		{
			float	rhw;
			if(fabsf(projected.w) > 0.0f)
			{
				rhw = 1.0f / ( projected.w );
			}
			else
			{
				rhw = 0.0f;
			}

			float eyespace_z = projected.z * rhw;
			float fFog = (eyespace_z * m_fFogMult) + m_fFogOffset;

			f32 fog_coeff = (f32)(fFog);
			f32 alpha = Clamp< f32 >( fog_coeff, 0.0f, 1.0f );

			// Set the alpha
			mVtxProjected[i].color.w = alpha;
		}

		if (m_bTextureEnable)
		{
			// Update texture coords n.b. need to divide tu/tv by bogus scale on addition to buffer

			// If the vert is already lit, then there is no normal (and hence we
			// can't generate tex coord)
			if (m_bTextureGen && m_bLightingEnable)
			{
				const Matrix4x4 & matWV = m_mModelView[m_dwModelViewTop];
				v3 & norm = vecTransformedNormal;

				// Assign the spheremap's texture coordinates
				mVtxProjected[i].t0.x = (0.5f * ( 1.0f + ( norm.x*matWV.m11 +
						                 norm.y*matWV.m21 +
										 norm.z*matWV.m31 ) ))/* * m_fTexScaleX*/;

				mVtxProjected[i].t0.y = (0.5f * ( 1.0f - ( norm.x*matWV.m12 +
						                 norm.y*matWV.m22 +
										 norm.z*matWV.m32 ) ))/* * m_fTexScaleY*/;
			}
			else
			{
				mVtxProjected[i].t0.x = (float)vert.tu * m_fTexScaleX;
				mVtxProjected[i].t0.y = (float)vert.tv * m_fTexScaleY;
			}

			//mVtxProjected[i].t1 = mVtxProjected[i].t0;
		}
/*
		DL_PF( "   %x: in(%f,%f,%f) -> out(%f,%f,%f,%f) trans(%f,%f,%f) tu/v(%f,%f)",
			i,
			w.x, w.y, w.z,
			projected.x, projected.y, projected.z, projected.w,
			mVtxProjected[i].pos.x, mVtxProjected[i].pos.y, mVtxProjected[i].pos.z,
			mVtxProjected[i].t0.x, mVtxProjected[i].t0.y );
*/
	}
}


//*****************************************************************************
// Assumes dwAddress has already been checked!	
// Don't inline - it's too big with the transform macros
// DKR seems to use longer vert info
//*****************************************************************************
void PSPRenderer::SetNewVertexInfoDKR(u32 dwAddress, u32 dwV0, u32 dwNum)
{
	s32 nFogR, nFogG, nFogB, nFogA;

	if (m_bFogEnable)
	{
		nFogR = COL_GETR(m_dwFogColor);
		nFogG = COL_GETG(m_dwFogColor);
		nFogB = COL_GETB(m_dwFogColor);
		nFogA = COL_GETA(m_dwFogColor);
	}
	s16 * pVtxBase = (s16*)(g_pu8RamBase + dwAddress);

	const Matrix4x4 matWorldProject = m_mModelView[m_dwModelViewTop] * m_mProjection[m_dwProjectionTop];

	u32 i;
	s32 nOff;

	nOff = 0;
	for (i = dwV0; i < dwV0 + dwNum; i++)
	{
		v4 w;
		u32 dwFlags;

		w.x = (float)pVtxBase[(nOff + 0) ^ 1];
		w.y = (float)pVtxBase[(nOff + 1) ^ 1];
		w.z = (float)pVtxBase[(nOff + 2) ^ 1];
		w.w = 1.0f;

		v4 & projected( mVtxProjected[i].ProjectedPos );
		projected = matWorldProject.Transform( w );	// Convert to w=1

		float	rhw;
		if(fabsf(projected.w) > 0.0f)
		{
			rhw = 1.0f / ( projected.w );
		}
		else
		{
			rhw = 1.0f;
		}

		dwFlags = 0;

		// Clip
		mVtxClipFlags[i] = dwFlags;

		s16 wA = pVtxBase[(nOff + 3) ^ 1];
		s16 wB = pVtxBase[(nOff + 4) ^ 1];

		s8 r = (s8)(wA >> 8);
		s8 g = (s8)(wA);
		s8 b = (s8)(wB >> 8);
		s8 a = (s8)(wB);

		v3 vecTransformedNormal;
		v4 colour;

		if (m_bLightingEnable)
		{
			v3 mn;		// modelnorm
			const Matrix4x4 & matWorld( m_mModelView[m_dwModelViewTop] );

			mn.x = (float)r; //norma_x;
			mn.y = (float)g; //norma_y;
			mn.z = (float)b; //norma_z;

			vecTransformedNormal = matWorld.TransformNormal( mn );
			vecTransformedNormal.Normalise();

			colour = LightVert(vecTransformedNormal);
			colour.w = a / 255.0f;
		}
		else
		{
			colour = v4( r / 255.0f, g / 255.0f, b / 255.0f, a / 255.0f );
		}

		// Assign true vert colour after lighting/fogging
		mVtxProjected[i].color = colour;

		/*if (m_bTextureEnable)
		{
			// Update texture coords n.b. need to divide tu/tv by bogus scale on addition to buffer
			v2 & t = m_vecTexture[i];

			// If the vert is already lit, then there is no normal (and hence we
			// can't generate tex coord)
			if (m_bTextureGen && m_bLightingEnable)
			{
				const Matrix4x4 & matWV = m_mModelView[m_dwModelViewTop];
				v3 & norm = vecTransformedNormal;

				// Assign the spheremap's texture coordinates
				t.x = (0.5f * ( 1.0f + ( norm.x*matWV.m00 +
										 norm.y*matWV.m10 +
										 norm.z*matWV.m20 ) ))/ * * m_fTexScaleX* /;

				t.y = (0.5f * ( 1.0f - ( norm.x*matWV.m01 +
										 norm.y*matWV.m11 +
										 norm.z*matWV.m21 ) ))/ * * m_fTexScaleY * /;
			}
			else
			{
				t.x = (float)vert.t.x * m_fTexScaleX;
				t.y = (float)vert.t.y * m_fTexScaleY;
			}
		}*/

		nOff += 5;
	}

}

v4	MAKE_COLF( u32 colour )
{
	v4 col( COL_GETR(colour) / 255.0f, COL_GETG(colour) / 255.0f, COL_GETB(colour) / 255.0f, COL_GETA(colour) / 255.0f );

	return col;
}

//*****************************************************************************
//
//*****************************************************************************
void PSPRenderer::SetVtxColor( u32 vert, u32 color )
{
	if ( vert < MAX_VERTS )
	{
		mVtxProjected[vert].color = MAKE_COLF( color );
	}
}

//*****************************************************************************
//
//*****************************************************************************
void PSPRenderer::SetVtxTextureCoord( u32 vert, short tu, short tv )
{
	if ( vert < MAX_VERTS )
	{
		mVtxProjected[vert].t0.x = (float)tu / 32.0f;
		mVtxProjected[vert].t0.y = (float)tv / 32.0f;
	}
}

//*****************************************************************************
//
//*****************************************************************************
void PSPRenderer::SetVtxXY( u32 vert, float x, float y )
{
	if ( vert < MAX_VERTS )
	{
		// XXX Needs reprojection?
		printf( "SetVtxXY\n" );
		mVtxProjected[vert].TransformedPos.x = x;
		mVtxProjected[vert].TransformedPos.y = y;
	}
}

//*****************************************************************************
//
//*****************************************************************************

static const int gAddressModeLookup[PSPRenderer::TAM_NUM_ADDRESS_MODES] =
{
	GU_CLAMP,	// TAM_CLAMP = 0,
	GU_CLAMP,	// TAM_MIRROR,
	GU_REPEAT,	// TAM_WRAP,
};

//*****************************************************************************
//
//*****************************************************************************
void PSPRenderer::SetAddressMode( u32 index, ETextureAddressMode mode_u, ETextureAddressMode mode_v )
{
	sceGuTexWrap(gAddressModeLookup[mode_u],gAddressModeLookup[mode_v]);
}

//*****************************************************************************
//
//*****************************************************************************
void PSPRenderer::SetAmbientLight( float r, float g, float b )
{
	mAmbientColour.x = r;
	mAmbientColour.y = g;
	mAmbientColour.z = b;
	//mAmbientColour.w = 1.0f;
}

//*****************************************************************************
//
//*****************************************************************************
void PSPRenderer::SetLightCol(u32 dwLight, u32 colour)
{
	mLights[dwLight].Colour.x = (float)((colour >> 24)&0xFF) / 255.0f;
	mLights[dwLight].Colour.y = (float)((colour >> 16)&0xFF) / 255.0f;
	mLights[dwLight].Colour.z = (float)((colour >>  8)&0xFF) / 255.0f;
	mLights[dwLight].Colour.w = 1.0f;	// Ignore light alpha
}

//*****************************************************************************
//
//*****************************************************************************
void PSPRenderer::SetLightDirection(u32 l, float x, float y, float z)
{
	v3		normal( x, y, z );
	normal.Normalise();

	mLights[l].Direction.x = normal.x;
	mLights[l].Direction.y = normal.y;
	mLights[l].Direction.z = normal.z;
	mLights[l].Padding0 = 0.0f;
}

//*****************************************************************************
// Init matrix stack to identity matrices
//*****************************************************************************
void PSPRenderer::ResetMatrices()
{
	Matrix4x4 mat;

	mat.SetIdentity();

	m_dwProjectionTop = 0;
	m_dwModelViewTop = 0;
	m_mProjection[0] = mat;
	m_mModelView[0] = mat;
}

//*****************************************************************************
//
//*****************************************************************************
void PSPRenderer::SetTextureFactor(u32 colour)
{
	if (m_dwTextureFactor != colour)
	{
		m_dwTextureFactor = colour;
		sceGuTexEnvColor(colour);
	}
}



//*****************************************************************************
//
//*****************************************************************************
void PSPRenderer::SetTexture( u32 index,
							CTexture* lpsTexture,
							s32  nTileLeft, s32 nTileTop,
							u32 dwTileWidth, u32 dwTileHeight )
{
	//lpsTexture = NULL; - For wireframe
	if (m_lpsTexture[index] != lpsTexture)
	{
		// Dereference the previous texture
		if ( m_lpsTexture[index] )
		{
			m_lpsTexture[index]->Release();
			m_lpsTexture[index] = NULL;
		}

		m_lpsTexture[index] = lpsTexture;

		if ( m_lpsTexture[index] )
		{
			m_lpsTexture[index]->AddRef();

			m_fTileWidth[index] = (float)dwTileWidth;
			m_fTileHeight[index] = (float)dwTileHeight;

			// Update the bogus scale parameters
			m_fTexLeft[index] = (float)nTileLeft;
			m_fTexTop[index]  = (float)nTileTop;

			m_fTexWidth[index] = (float)m_lpsTexture[index]->GetWidth();
			m_fTexHeight[index] = (float)m_lpsTexture[index]->GetHeight();
			
			m_fInvTileWidth[index] = 1.0f / m_fTileWidth[index];
			m_fInvTileHeight[index] = 1.0f / m_fTileHeight[index];
			m_fInvTexWidth[index] = 1.0f / m_fTexWidth[index];
			m_fInvTexHeight[index] = 1.0f / m_fTexHeight[index];

		}

		gTextureChanged = true;
	}
}

//*****************************************************************************
//
//*****************************************************************************
// Generate texture coords?
void PSPRenderer::SetTextureGen(bool bTextureGen)
{
	m_bTextureGen = bTextureGen;
}

//*****************************************************************************
//
//*****************************************************************************
void PSPRenderer::SetTextureScale(float fScaleX, float fScaleY)
{
	m_fTexScaleX = fScaleX;
	m_fTexScaleY = fScaleY;
}

//*****************************************************************************
//
//*****************************************************************************
void	PSPRenderer::SetCullMode( bool bCullFront, bool bCullBack )
{
	m_bCullFront = bCullFront;
	m_bCullBack = bCullBack;
}

//*****************************************************************************
//
//*****************************************************************************
void PSPRenderer::SetAlphaRef(u32 dwAlpha)
{
	if (m_dwAlpha != dwAlpha)
	{
		m_dwAlpha = dwAlpha;
		sceGuAlphaFunc(GU_GEQUAL, dwAlpha, 0xff );
	}
}

//*****************************************************************************
//
//*****************************************************************************
void	PSPRenderer::SetScissor( s32 x0, s32 y0, s32 x1, s32 y1 )
{
	s32	x = s32(x0 * m_fScreenMultX);
	s32 y = s32(y0 * m_fScreenMultY);
	s32 w = s32((x1-x0) * m_fScreenMultX);
	s32 h = s32((y1-y0) * m_fScreenMultY);

	sceGuScissor(x,y, x+w,y+h);
}

//*****************************************************************************
//
//*****************************************************************************
void PSPRenderer::SetProjection(const Matrix4x4 & mat, bool bPush, EMatrixLoadStyle load_style) 
{
	if (bPush)
	{
		if (m_dwProjectionTop >= (MATRIX_STACK_SIZE-1))
			DBGConsole_Msg(0, "Pushing past proj stack limits!");
		else
			m_dwProjectionTop++;

		if (load_style == MATRIX_LOAD)
			// Load projection matrix
			m_mProjection[m_dwProjectionTop] = mat;
		else
			m_mProjection[m_dwProjectionTop] = mat * m_mProjection[m_dwProjectionTop-1];
		
	}
	else
	{
		if (load_style == MATRIX_LOAD)
			// Load projection matrix
			m_mProjection[m_dwProjectionTop] = mat;
		else
			m_mProjection[m_dwProjectionTop] = mat * m_mProjection[m_dwProjectionTop];

	}

}

//*****************************************************************************
//
//*****************************************************************************
void PSPRenderer::SetWorldView(const Matrix4x4 & mat, bool bPush, EMatrixLoadStyle load_style)
{
	// ModelView
	if (bPush)
	{
		if (m_dwModelViewTop >= (MATRIX_STACK_SIZE-1))
			DBGConsole_Msg(0, "Pushing past modelview stack limits!");
		else
			m_dwModelViewTop++;

		// We should store the current projection matrix...
		if (load_style == MATRIX_LOAD)
		{
			// Load projection matrix
			m_mModelView[m_dwModelViewTop] = mat;
		}
		else			// Multiply projection matrix
		{
			m_mModelView[m_dwModelViewTop] = mat * m_mModelView[m_dwModelViewTop-1];
		}
	}
	else	// NoPush
	{
		if (load_style == MATRIX_LOAD)
		{
			// Load projection matrix
			m_mModelView[m_dwModelViewTop] = mat;
		}
		else
		{
			// Multiply projection matrix
			m_mModelView[m_dwModelViewTop] = mat * m_mModelView[m_dwModelViewTop];
		}

	}
}

//*****************************************************************************
//
//*****************************************************************************
void PSPRenderer::PopProjection()
{
	if (m_dwProjectionTop > 0)
		m_dwProjectionTop--;
	//else
	//	DBGConsole_Msg(0, "Popping past projection stack limits");
}

//*****************************************************************************
//
//*****************************************************************************
void PSPRenderer::PopWorldView()
{
	if (m_dwModelViewTop > 0)
		m_dwModelViewTop--;
	//else
	//	DBGConsole_Msg(0, "Popping past worldview stack limits");
}

//*****************************************************************************
//
//*****************************************************************************

void PSPRenderer::PrintActive()
{
#ifdef DAEDALUS_DEBUG_DISPLAYLIST
	if (gDisplayListFile != NULL)
	{
		const Matrix4x4 mat( m_mModelView[m_dwModelViewTop] * m_mProjection[m_dwProjectionTop] );
		DL_PF(
			" %#+12.5f %#+12.5f %#+12.5f %#+12.5f\n"
			" %#+12.5f %#+12.5f %#+12.5f %#+12.5f\n"
			" %#+12.5f %#+12.5f %#+12.5f %#+12.5f\n"
			" %#+12.5f %#+12.5f %#+12.5f %#+12.5f\n",
			mat.m[0][0], mat.m[0][1], mat.m[0][2], mat.m[0][3],
			mat.m[1][0], mat.m[1][1], mat.m[1][2], mat.m[1][3],
			mat.m[2][0], mat.m[2][1], mat.m[2][2], mat.m[2][3],
			mat.m[3][0], mat.m[3][1], mat.m[3][2], mat.m[3][3]);
	}
#endif
}
